/*------------------------------------------------------------------------------
* pntpos.c : single point positioning
*
*          Copyright (C) 2006 by T.TAKASU, All rights reserved.
*
* version : $Revision: 1.2 $ $Date: 2006/12/12 23:27:39 $
* history : 2006/11/28 0.1 new
*-----------------------------------------------------------------------------*/
#include <stdlib.h>
#include "rtkcmn.h"
#include "pntpos.h"

const static char rcsid[]="$Id: pntpos.c,v 1.2 2006/12/12 23:27:39 ttaka Exp ttaka $";

#define MAXITR		10			/* max number of iteration */

/* point positioning for an epoch ----------------------------------------------
* compute receiver position and clock-bias by single point positioning
* args   : obsd_t *obs		I	observation data records
*		   int	  n			I	number of observation data records
*		   double *rs		I	satellilte positions (ecef) (m)
*		   double *dts		I	satellilte clock-biases (sec)
*		   double *ion		I	broadcast ionosphere model parameters
*		   double elmin		I	elevation cutoff angle (deg)
*		   double *rr		O	receiver position (ecef) (m)
*		   double *dtr		O	receiver clock-bias (sec)
*		   double *azel		O	azimuth/elevation angle {az,el} (rad)
* return : status (0:ok,-1:divergent,-2:least square error)
* note   : rs[(0:2)+i*3],dts[i] = obs[i] satellite position/clock-bias
*          azel[(0:1)+i*2] = obs[i] azimuth/elevation angle
*-----------------------------------------------------------------------------*/
extern int pointp(const obsd_t *obs, int n, const double *rs, const double *dts,
				  const double *ion, double elmin, double *rr, double *dtr,
				  double *azel)
{
	int i,j,k,m;
	double r,pos[3],e[3],H[4*MAXOBS],x[4],y[MAXOBS];
	
	trace_("pointp()\n");
	
	for (i=0;i<3;i++) rr[i]=0.0; *dtr=0.0; /* initial position/clock-bias */
	
	for (i=0;i<MAXITR;i++) {
		
		ecef2pos(rr,pos); /* receiver geodetic postion */ 
		
		for (j=m=0;j<n&&j<MAXOBS;j++) {
			azel[j*2]=azel[1+j*2]=0.0;
			
			/* geometric distance */
			if (ISNAN(obs[j].P[0])||(r=geodist(rs+j*3,rr,e))<0.0) continue;
			
			/* azimuth/elevation angle */
			satazel(pos,e,azel+j*2); if (azel[1+j*2]<elmin*D2R) continue;
			
			/* pseudorange model */
			r+=C*(*dtr-dts[j])+ionmodel(obs[j].time,ion,pos,azel+j*2)+
			   tropmodel(pos,azel+j*2);
			
			/* design matrix */
			for (k=0;k<4;k++) H[k+4*m]=k<3?-e[k]:1.0;
			
			/* observation - model */
			y[m++]=obs[j].P[0]-r;
		}
		/* least square estimation */
		if (lsq(H,y,4,m,x)) return -2;
		
		for (j=0;j<3;j++) rr[j]+=x[j]; *dtr+=x[3]/C;
		
		if (NORM(x)<1E-4) return 0;
	}
	return -1;
}
/* point positioning for all epochs --------------------------------------------
* input obs/nav data, compute receiver positions and output result
* args   : infunc_t input 	I 	data input callback
*		   outfunc_t output	I	data output callback
*		   double elmin		I	elevation cutoff angle (deg)
* return : none
* note   : data input and output callback shall comply the following prototypes.
*            int func(obsd_t *obs, nav_t *nav);
*            void func(gtime_t t, double *rr);
*-----------------------------------------------------------------------------*/
extern void pntpos(infunc_t input, outfunc_t output, double elmin)
{
	int n;
	double rs[3*MAXOBS],dts[MAXOBS],azel[2*MAXOBS],rr[3],dtr;
	obsd_t *obs;
	nav_t *nav;
	
	trace_("pntpos()\n");
	
	while ((n=(*input)(&obs,&nav))>0) { /* input callback */
		
		/* satellite position/clock-bias */
		satpos(obs,n,nav,rs,dts);
		
		/* point positioning */
		if (pointp(obs,n,rs,dts,nav->ion,elmin,rr,&dtr,azel)) continue;
		
		(*output)(timeadd(obs[0].time,-dtr),rr); /* output callback */
	}
}
