/*------------------------------------------------------------------------------
* rtkcmn.c : rtk common functions
*
*          Copyright (C) 2006 by T.TAKASU, All rights reserved.
*
* version : $Revision: 1.2 $ $Date: 2006/12/12 23:27:39 $
* history : 2006/05/01 0.1 new
*-----------------------------------------------------------------------------*/
#include <stdio.h>
#include <string.h>
#include <time.h>
#include "rtkcmn.h"

const static char rcsid[]="$Id: rtkcmn.c,v 1.2 2006/12/12 23:27:39 ttaka Exp ttaka $";

#define MAXDTOE		43200.0		/* max time difference to toe (sec) */

const static double gpst0[]={1980,1,6,0,0,0};	/* gpsweek 0 - day 0 */

/* convert epoch to gtime_t ----------------------------------------------------
* convert epoch to gtime_t struct
* args   : double *ep		I	epoch {year,month,day,hour,min,sec}
* return : gtime_t struct
*-----------------------------------------------------------------------------*/
extern gtime_t epoch2time(const double *ep)
{
	struct tm t={(int)ep[5],(int)ep[4],(int)ep[3],(int)ep[2],(int)ep[1]-1,
				 (int)ep[0]-1900,0,0,0};
	gtime_t time={mktime(&t),ep[5]-(int)ep[5]};
	return time;
}
/* time to epoch ---------------------------------------------------------------
* convert gtime_t struct to epoch
* args   : gtime_t t		I	gtime_t struct
*          double *ep		O	epoch {year,month,day,hour,min,sec}
* return : none
*-----------------------------------------------------------------------------*/
extern void time2epoch(gtime_t t, double *ep)
{
	struct tm *tt=localtime(&t.time);
	ep[0]=tt->tm_year+1900; ep[1]=tt->tm_mon+1; ep[2]=tt->tm_mday;
	ep[3]=tt->tm_hour; ep[4]=tt->tm_min; ep[5]=tt->tm_sec+t.sec;
}
/* gpstime to time -------------------------------------------------------------
* convert gpstime to gtime_t struct
* args   : int    week		I	gps week number
*          double sec		I	gpstime (time of gps week) (sec)
* return : gtime_t struct
*-----------------------------------------------------------------------------*/
extern gtime_t gpst2time(int week, double sec)
{
	gtime_t t=epoch2time(gpst0);
	t.time+=86400*7*week+(int)sec;
	t.sec=sec-(int)sec;
	return t;
}
/* time to gpstime -------------------------------------------------------------
* convert gtime_t struct to gpstime
* args   : gtime_t t		I	gtime_t struct
*          int    *week		O	gps week number
* return : gpstime (time of gps week) (sec)
*-----------------------------------------------------------------------------*/
extern double time2gpst(gtime_t t, int *week)
{
	gtime_t t0=epoch2time(gpst0);
	long sec=t.time-t0.time;
	*week=(int)(sec/(86400*7));
	return (double)(sec-(long)*week*86400*7)+t.sec;
}
/* add time --------------------------------------------------------------------
* add time to gtime_t struct
* args   : gtime_t t		I	gtime_t struct
*          double sec		I	time to add (sec)
* return : gtime_t struct (t+sec)
*-----------------------------------------------------------------------------*/
extern gtime_t timeadd(gtime_t t, double sec)
{
	double tt;
	t.sec+=sec; tt=floor(t.sec); t.time+=(int)tt; t.sec-=tt;
	return t;
}
/* time difference -------------------------------------------------------------
* difference between gtime_t structs
* args   : gtime_t t1,t2	I	gtime_t structs
* return : time difference (t1-t2) (sec)
*-----------------------------------------------------------------------------*/
extern double timediff(gtime_t t1, gtime_t t2)
{
	return difftime(t1.time,t2.time)+t1.sec-t2.sec;
}
/* time to string --------------------------------------------------------------
* convert gtime_t struct to string
* args   : gtime_t t		I	gtime_t struct
*		   char   *s		IO	string ("YYYY/MM/dd hh:mm:ss....")
*		   int	  n			I	columns under decimal point
* return : none
*-----------------------------------------------------------------------------*/
extern void time2str(gtime_t t, char *s, int n)
{
	double ep[6];
	if (1.0-t.sec<0.5/pow(10.0,n)) {t.time++; t.sec=0.0;};
	time2epoch(t,ep);
	sprintf(s,"%04.0f/%02.0f/%02.0f %02.0f:%02.0f:%0*.*f",ep[0],ep[1],ep[2],
		    ep[3],ep[4],n<=0?2:n+3,n,ep[5]);
}
/* string to time --------------------------------------------------------------
* convert substring to gtime_t struct
* args   : char   *s		I	string ("... YYYY MM dd hh mm ss ...")
*		   int	  i,n		I	substring position and width
*          gtime_t t		O	gtime_t struct
* return : status (0:ok,0>:error)
*-----------------------------------------------------------------------------*/
extern int str2time(const char *s, int i, int n, gtime_t *t)
{
	char str[64];
	double ep[6];
	if (strlen(s)<i) return 0;
	strncpy(str,s+i,n); str[n]='\0';
	if (sscanf(str,"%lf %lf %lf %lf %lf %lf",ep,ep+1,ep+2,ep+3,ep+4,ep+5)<6)
		return -1;
	if (ep[0]<100.0) ep[0]+=ep[0]<80.0?2000.0:1900.0;
	*t=epoch2time(ep);
	return 0;
}
/* string to number ------------------------------------------------------------
* convert substring to number
* args   : char   *s		I	string ("... nnn.nnn ...")
*		   int	  i,n		I	substring position and width
* return : converted number (NAN:error)
*-----------------------------------------------------------------------------*/
extern double str2num(const char *s, int i, int n)
{
	char str[64],*p=str;
	double value;
	if (strlen(s)<i) return NAN;
	for (s+=i;*s&&--n>=0;s++) *p++=*s=='D'?'E':*s;
	*p='\0';
	return sscanf(str,"%lf",&value)==1?value:NAN;
}
/* transform ecef to geodetic postion ------------------------------------------
* transform ecef position to geodetic position
* args   : double *r		I	ecef position {x,y,z} (m)
* 		   double *pos		O	geodetic position {lat,lon,h} (rad,m)
* return : none
*-----------------------------------------------------------------------------*/
extern void ecef2pos(const double *r, double *pos)
{
	double e2=FE*(2.0-FE),r2=SQR(r[0])+SQR(r[1]);
	double z,zk,v=RE,sinp;
	for (z=r[2],zk=0.0;ABS(z-zk)>=1E-4;) {
		zk=z;
		sinp=z/sqrt(r2+SQR(z));
		v=RE/sqrt(1.0-e2*SQR(sinp));
		z=r[2]+v*e2*sinp;
	}
	pos[0]=r2<1E-12?PI/2.0:atan(z/sqrt(r2));
	pos[1]=atan2(r[1],r[0]);
	pos[2]=sqrt(r2+SQR(z))-v;
}
/* transform geodetic to ecef --------------------------------------------------
* transform geodetic position to ecef position
* args   : double *pos		I	geodetic position {lat,lon,h} (rad,m)
*		   double *r		O	ecef position {x,y,z} (m)
* return : none
*-----------------------------------------------------------------------------*/
extern void pos2ecef(const double *pos, double *r)
{
	double e2=FE*(2.0-FE);
	double sinp=sin(pos[0]),cosp=cos(pos[0]);
	double sinl=sin(pos[1]),cosl=cos(pos[1]);
	double v=RE/sqrt(1.0-e2*SQR(sinp));
	r[0]=(v+pos[2])*cosp*cosl;
	r[1]=(v+pos[2])*cosp*sinl;
	r[2]=(v*(1.0-e2)+pos[2])*sinp;
}
/* transform to local tangental coordinate -------------------------------------
* transform ecef vector to local tangental coordinate
* args   : double *pos		I	geodetic position {lat,lon,h} (rad,m)
*		   double *r		I	vector in ecef coordinate {x,y,z}
*		   double *e		O	vector in local tangental coordinate {e,n,u}
* return : none
*-----------------------------------------------------------------------------*/
extern void tolocal(const double *pos, const double *r, double *e)
{
	double sinp=sin(pos[0]),cosp=cos(pos[0]);
	double sinl=sin(pos[1]),cosl=cos(pos[1]);
	e[0]=-sinl*r[0]+cosl*r[1];
	e[1]=-sinp*cosl*r[0]-sinp*sinl*r[1]+cosp*r[2];
	e[2]= cosp*cosl*r[0]+cosp*sinl*r[1]+sinp*r[2];
}
/* navigation message to satellite position/clock-bias -----------------------*/
static void nav2pos(gtime_t t, const navm_t *nav, double pr, double *rs,
				    double *dts)
{
	double tk,tc,tt,M,E,Ek,phi,sin2p,cos2p,r,i,x,y,OMG;
	
	tk=timediff(t,nav->toe);
	tc=timediff(t,nav->toc);
	tt=pr/C+(nav->f0+nav->f1*tc+nav->f2*SQR(tc));
	tk-=tt; tc-=tt; /* satellite signal transmition time */
	M=nav->M0+(sqrt(MU/SQR(nav->A)/nav->A)+nav->deln)*tk;
	for (E=M,Ek=0.0;ABS(E-Ek)>1E-12;) {
		Ek=E; E=M+nav->e*sin(Ek);
	}
	phi=atan2(sqrt(1.0-SQR(nav->e))*sin(E),cos(E)-nav->e)+nav->omg;
	sin2p=sin(2.0*phi);
	cos2p=cos(2.0*phi);
	phi+=nav->cus*sin2p+nav->cuc*cos2p;
	r=nav->A*(1.0-nav->e*cos(E))+nav->crs*sin2p+nav->crc*cos2p;
	i=nav->i0+nav->cis*sin2p+nav->cic*cos2p+nav->idot*tk;
	OMG=nav->OMG0+(nav->OMGd-OMGE)*tk-OMGE*nav->toes;
	x=r*cos(phi);
	y=r*sin(phi);
	rs[0]=x*cos(OMG)-y*cos(i)*sin(OMG);
	rs[1]=x*sin(OMG)+y*cos(i)*cos(OMG);
	rs[2]=y*sin(i);
	*dts=nav->f0+nav->f1*tc+nav->f2*SQR(tc);
	*dts-=2.0*sqrt(MU*nav->A)*nav->e*sin(E)/SQR(C)+nav->tgd;
}
/* satellite positions/clock-biases --------------------------------------------
* compute satellite positions and clock-biases from navigation messages
* args   : obsd_t *obs		I	observation data records
*		   int	   n		I	number of observation data records
*		   nav_t  *nav		I	navigation messages
*		   double *rs		O	satellilte positions (ecef) (m)
*		   double *dts		O	satellilte clock-bias (sec)
* return : none
* note   : rs[(0:2)+i*3],dts[i] = obs[i] satellite position/clock-bias
*		   if no navigation messages, set 0.0 to rs, dts
*-----------------------------------------------------------------------------*/
extern void satpos(const obsd_t *obs, int n, const nav_t *nav, double *rs,
				   double *dts)
{
	int i,j,k;
	double tt,tmin;
	
	trace_("satpos()\n");
	
	for (i=0;i<n&&i<MAXOBS;i++) {
		/* search navigation message tow closest to first epoch */
		for (j=k=0,tmin=1E99;j<nav->n;j++) {
			if (nav->msg[j].sat!=obs[i].sat) continue;
			tt=timediff(obs[0].time,nav->msg[j].toe);
			if (ABS(tt)<tmin) {k=j; tmin=ABS(tt);}
		}
		/* satellite position/clock-bias */
		if (tmin<=MAXDTOE&&nav->msg[k].svh==0) {
			nav2pos(obs[i].time,nav->msg+k,obs[i].P[0],rs+i*3,dts+i);
		}
		else rs[i*3]=rs[1+i*3]=rs[2+i*3]=dts[i]=0.0;
	}
}
/* geometric distance ----------------------------------------------------------
* compute geometric distance and receiver-to-satellite unit vector
* args   : double *rs		I	satellilte position (ecef at transmition) (m)
*		   double *rr		I	receiver position (ecef at reception) (m)
*		   double *e		O	receiver-to-satellilte unit vector
* return : geometric distance (m) (<0:error/no satellite position)
*-----------------------------------------------------------------------------*/
extern double geodist(const double *rs, const double *rr, double *e)
{
	int i,j;
	double r,rk,cost,sint;
	
	trace_("geodist()\n");
	
	if (NORM(rs)<RE) return -1.0;
	for (i=0,r=2E7,rk=0.0;i<10;i++,rk=r) {
		cost=cos(OMGE*r/C);
		sint=sin(OMGE*r/C);
		e[0]= cost*rs[0]+sint*rs[1]-rr[0];
		e[1]=-sint*rs[0]+cost*rs[1]-rr[1];
		e[2]=rs[2]-rr[2];
		r=NORM(e);
		if (ABS(r-rk)<1E-4) {
			for (j=0;j<3;j++) e[j]/=r;
			return r;
		}
	}
	return -1.0;
}
/* satellite azimuth/elevation angle -------------------------------------------
* compute satellite azimuth/elevation angle
* args   : double *pos		I	geodetic position {lat,lon,h} (rad/m)
*		   double *e		I	receiver-to-satellilte unit vevtor
*		   double *azel		O	azimuth/elevation angle {az,el} (rad)
* return : none
*-----------------------------------------------------------------------------*/
extern void satazel(const double *pos, const double *e, double *azel)
{
	double ee[3];
	
	trace_("satazel()\n");
	
	if (pos[2]<=-RE) {
		azel[0]=0.0;
		azel[1]=PI/2.0;
		return;
	}
	tolocal(pos,e,ee);
	azel[0]=atan2(ee[0],ee[1]);
	azel[1]=asin(ee[2]);
}
/* ionospheric model -----------------------------------------------------------
* compute ionospheric delay by broadcast ionosphere model (klobuchar model)
* args   : gtime_t t		I	gtime_t struct
*		   double *ion		I	iono model parameters {a0,a1,a2,a3,b0,b1,b2,b3}
*		   double *pos		I	receiver position {lat,lon,h} (rad,m)
*		   double *azel		I	azimuth/elevation angle {az,el} (rad)
* return : L1 ionospheric delay (m)
*-----------------------------------------------------------------------------*/
extern double ionmodel(gtime_t t, const double *ion, const double *pos,
					   const double *azel)
{
	int week;
	double tt,f,psi,phi,lam,amp,per,x;
	
	f=1.0+16.0*pow(0.53-azel[1]/PI,3.0);
	psi=0.0137/(azel[1]/PI+0.11)-0.022;
	phi=pos[0]/PI+psi*cos(azel[0]);
	phi=MAX(phi,-0.416); phi=MIN(phi,0.416);
	lam=pos[1]/PI+psi*sin(azel[0])/cos(phi*PI);
	phi+=0.064*cos((lam-1.617)*PI);
	amp=ion[0]+phi*(ion[1]+phi*(ion[2]+phi*ion[3]));
	per=ion[4]+phi*(ion[5]+phi*(ion[6]+phi*ion[7]));
	amp=MAX(amp,0.0); per=MAX(per,72000.0);
	tt=fmod(43200.0*lam+time2gpst(t,&week),86400.0); if (tt<0.0) tt+=86400.0;
	x=2.0*PI*(tt-50400.0)/per;
	return ABS(x)<1.57?C*f*(5E-9+amp*(1.0+SQR(x)*(-0.5+SQR(x)/24.0))):C*f*5E-9;
}
/* tropospheric model ----------------------------------------------------------
* compute tropospheric delay by standard atmosphere and saastamoinen model
* args   : double *pos		I	receiver position {lat,lon,h} (rad,m)
*		   double *azel		I	azimuth/elevation angle {az,el} (rad)
* return : tropospheric delay (m)
*-----------------------------------------------------------------------------*/
extern double tropmodel(const double *pos, const double *azel)
{
	double h,pres,temp,e,humi=50.0;
	
	if (pos[2]>1E4||azel[1]<=0) return 0.0;
	h=MAX(pos[2],0.0);
	pres=1013.25*pow(1.0-2.2557E-5*h,5.2568);
	temp=15.0-6.5E-3*h+273.15;
	e=6.108*exp((17.15*temp-4684.0)/(temp-38.45))*humi/100.0;
	return 0.002277/sin(azel[1])*(pres+(1255.0/temp+0.05)*e);
}
/* least square estimation -----------------------------------------------------
* estimate parameters by least square adustment (x=H'\y)
* args   : double *H		IO	transpose of design matrix (n x m)
*		   double *y		IO	measurement vector (m x 1)
*		   int    n			I	number of parameters
*		   int    m			I	number of measurements
*		   double *x		O	estimated parameter vector (n x 1)
* return : status (0:ok,<0:error)
*-----------------------------------------------------------------------------*/
extern int lsq(double *H, double *y, int n, int m, double *x)
{
	int k=1,lwork=n*16,info=-1;
	double *work=MAT(lwork);
	
	if (work&&m>=n) {
		dgels_("T",&n,&m,&k,H,&n,y,&m,work,&lwork,&info);
		memcpy(x,y,sizeof(double)*n);
	}
	free(work);
	return info;
}
/* filter states ---------------------------------------------------------------
* update state vector and covariances by kalman filter measurment update
* args   : double *x		IO	state vector (n x 1)
*		   double *P		IO	covariance matrix of state vector (n x n)
*		   double *H		I	transpose of design matrix (n x m)
*		   double *v		I	innovation (measurement - model) (m x 1)
*		   double *R		I	covariance matrix of measurement error (m x m)
*		   int    n			I	number of parameters
*		   int    m			I	number of measurements
* return : status (0:ok,<0:error)
*-----------------------------------------------------------------------------*/
extern int filter(double *x, double *P, double *H, double *v, const double *R,
				  int n, int m)
{
	int k=1,info;
	double one=1.0,min=-1.0,zero=0.0;
	double *PH=MAT(n*m),*Q=MAT(m*m),*K=MAT(n*m);
	
	if (!PH||!Q||!K) {free(PH); free(Q); free(K); return -1;}
	
	memcpy(Q,R,sizeof(double)*m*m);
	dsymm_("L","U",&n,&m,&one,P,&n,H,&n,&zero,PH,&n);		/* PH=P*H */
	dgemm_("T","N",&m,&m,&n,&one,H,&n,PH,&n,&one,Q,&m);		/* Q=H'*PH+R */
	dpotrf_("U",&m,Q,&m,&info);								/* Q=inv(Q) */
	if (!info) dpotri_("U",&m,Q,&m,&info);
	if (!info) {
		dsymm_("R","U",&n,&m,&one,Q,&m,PH,&n,&zero,K,&n);	/* K=PH*Q */
		dgemm_("N","N",&n,&k,&m,&one,K,&n,v,&m,&one,x,&n);	/* x=x+K*v */
		dgemm_("N","T",&n,&n,&m,&min,K,&n,PH,&n,&one,P,&n);	/* P=P-K*PH' */
	}
	free(PH); free(Q); free(K);
	return info;
}
/* transform state vector ------------------------------------------------------
* transform state vector and covariance matrix (y=A*x; Q=A*P*A')
* args   : double *x		I	state vector (n x 1)
*		   double *P		I	covariance matrix of state vector (n x n)
*		   double *A		I	transformation matrix (m x n)
*		   int    n, m		I	number of states and transformed states
*		   double *y		O	transformed state vector (m x 1)
*		   double *Q		O	transformed covariance matrix (m x m)
* return : status (0:ok,<0:error)
*-----------------------------------------------------------------------------*/
extern void transx(const double *x, const double *P, const double *A, int n,
				   int m, double *y, double *Q)
{
	int k=1;
	double one=1.0,zero=0.0,*AP=MAT(m*n);
	
	if (!AP) return;
	dgemm_("N","N",&m,&k,&n,&one,A,&m,x,&n,&zero,y,&m);		/* y=A*x */
	dgemm_("N","N",&m,&n,&n,&one,A,&m,P,&n,&zero,AP,&m);	/* AP=A*P */
	dgemm_("N","T",&m,&m,&n,&one,AP,&m,A,&m,&zero,Q,&m);	/* Q=AP*A' */
	free(AP);
}
/* print matrix  -------------------------------------------------------------*/
extern void printm(const double *A, int n, int m, int p, int q)
{
	int i,j;
	for (i=0;i<n;i++) {
		for (j=0;j<m;j++) printf("%*.*f ",p,q,A[i+j*n]);
		printf("\n");
	}
}
